__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from citelang.main import Client


def main(args, parser, extra, subparser):

    cli = Client(quiet=args.quiet, settings_file=args.settings_file)

    if args.clear:
        cli.clear_cache(force=args.force)
    else:
        print(cli.settings.cache_dir)



#!/usr/bin/python

import pytest
import os

import citelang.utils as utils
import citelang.main.client as client
import citelang.main.schemas as schemas


@pytest.mark.parametrize(
    "manager,name",
    [
        ("pypi", "requests"),
        ("spack", "caliper"),
        ("github", "singularityhub/singularity-cli"),
    ],
)
def test_package(manager, name, tmp_path):
    """
    Test packages
    """
    cli = client.Client(quiet=False)
    result = cli.package(name=name, manager=manager)
    for prop in schemas.package["required"]:
        assert prop in result.data
    result.print_json()

    outfile = os.path.join(str(tmp_path), "%s.json" % os.path.basename(name))
    result.save(outfile)
    loaded = utils.read_json(outfile)
    assert loaded == result.data
    result.table()


def test_list(tmp_path):
    """
    Test listing package managers
    """
    cli = client.Client(quiet=False)
    result = cli.package_managers()

    # 30 at time of parsing
    assert len(result.data) > 30
    result.print_json()

    outfile = os.path.join(str(tmp_path), "package_managers.json")
    result.save(outfile)
    loaded = utils.read_json(outfile)
    assert loaded == result.data
    result.table()


@pytest.mark.parametrize(
    "manager,name",
    [
        ("pypi", "requests"),
        ("spack", "caliper"),
        ("github", "singularityhub/singularity-cli"),
    ],
)
def test_credit(manager, name, tmp_path):
    """
    Test credit function
    """
    cli = client.Client(quiet=False)
    result = cli.credit(name=name, manager=manager)
    result.print_json()

    outfile = os.path.join(str(tmp_path), "%s.json" % os.path.basename(name))
    result.save(outfile)
    loaded = utils.read_json(outfile)
    assert loaded == result.data
    result.print_result()


@pytest.mark.parametrize("graphtype", [(None), ("cypher"), ("gexf"), ("dot")])
def test_graph(graphtype):
    cli = client.Client(quiet=False)
    cli.graph(name="requests", manager="pypi", fmt=graphtype)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"


import secrets
import string
import sys


class GraphBase:
    def __init__(self, data, outfile=None):
        self.data = data
        self.uids = {}
        self.deps_ids = {}
        self.parse()
        self._outfile = outfile

    @property
    def outfile(self):
        if not self._outfile:
            self._outfile = sys.stdout
        return self._outfile

    def parse(self):
        """
        Create a flattened list of dependency names, and generate placeholder names
        """
        next_nodes = [self.data]
        while next_nodes:
            next_node = next_nodes.pop(0)
            if next_node.name not in self.uids:
                self.uids[next_node.name] = self.generate_placeholder()
            for child in next_node.children:
                next_nodes.append(child)

    def iter_nodes(self):
        """
        Iterate nodes, yielding node, weight, and indentation
        """
        pass

    def generate_placeholder(self):
        """
        Generate a unique placeholder name for a node.
        """
        # Taken from the Python3 documentation:
        # https://docs.python.org/3/library/secrets.html#recipes-and-best-practices
        while True:
            name = "".join(
                secrets.choice(string.ascii_letters) for i in range(8)
            ).lower()
            if name not in self.uids:
                return name



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

# Custom package managers not in libraries IO

from .github import GitHubManager
from .spack import SpackManager

# Registered endpoints (populated on init)
managers = {}
manager_names = []

for manager in [GitHubManager, SpackManager]:
    manager_names.append(manager.name)
    managers[manager.name] = manager

manager_names.sort()



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from rich import print
from rich.text import Text
from rich.tree import Tree
import random

# Randomly select an AMAZING icon
# Note some of these require TWO spaces after the icon to render properly!
# The faded heart is my favorite, so we want it 1/2 the time
icons = ["⭐️ ", "✨️ ", "💛️ ", "🔷️ ", "❤️  ", "💜️ ", "🧡️ ", "🔴️ ", "♦️  ", "🔶️ "] + [
    "💗️ "
] * 10
root_icons = ["🦄️", "☕️", "🤓️", "⭐️"]


def print_tree(root):

    # Randomly select icons
    icon = random.choice(icons)
    root_icon = random.choice(root_icons)

    node = Text(f"{root_icon} {root.name}", "bold magenta")
    credit = round(root.credit, root.round_by)
    node.append(f" ({credit})", "blue")
    tree = Tree(
        node,
        guide_style="bold bright_blue",
    )

    generate_tree(root, tree=tree, icon=icon)
    print(tree)


def generate_tree(next_node, icon, tree=None) -> None:
    """
    Recursively build a Tree with a dependency result
    """
    for child in next_node.children:
        # We won't have a tree on the first run
        credit = round(child.credit, child.round_by)
        node = Text(child.name, "green")
        node.highlight_regex(r"\..*$", "bold red")
        node.stylize(f"link file://{child.name}")
        node.append(f" ({credit})", "blue")
        branch = tree.add(Text(icon) + node)
        generate_tree(child, tree=branch, icon=icon)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from citelang.main import Client
import citelang.utils as utils


def main(args, parser, extra, subparser):

    cli = Client(quiet=args.quiet, settings_file=args.settings_file)
    result = cli.render(
        args.filename,
        use_cache=not args.no_cache,
        max_depth=args.max_depth,
        max_deps=args.max_depth,
        min_credit=args.min_credit,
        credit_split=args.credit_split,
    )

    if args.outfile:
        utils.write_file(result.render(), args.outfile)
        print("Saved to %s" % args.outfile)
    else:
        print(result.render())



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from citelang.logger import logger
import citelang.utils as utils
import citelang.main.endpoints as endpoints
import citelang.main.result as results
from citelang.main.settings import Settings

import time

import os
import json
import requests
import shutil


class BaseClient:
    """
    A baseclient controls interactions with endpoints and the cache.
    """

    def __init__(self, settings_file=None, validate=True, quiet=False, **kwargs):
        self.quiet = quiet
        self.session = requests.session()
        self.headers = {"Accept": "application/json", "User-Agent": "citelang-python"}
        self.params = {"per_page": 100}
        self.getenv()

        # keep a cache of data for a session
        self._cache = {}

        # If we don't have default settings, load
        if not hasattr(self, "settings"):
            self.settings = Settings(settings_file, validate=validate)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "[citelang-client]"

    def getenv(self):
        """
        Get any token / username set in the environment
        """
        self.api_key = os.environ.get("CITELANG_LIBRARIES_KEY")
        if self.api_key:
            self.params.update({"api_key": self.api_key})

    def clear_cache(self, force=False):
        """
        Clear the cache (with confirmation).
        """
        if not force and not utils.confirm_action(
            "Are you sure you want to clear the cache? "
        ):
            return
        if os.path.exists(self.settings.cache_dir):
            shutil.rmtree(self.settings.cache_dir)

    def cache(self, name, result):
        """
        Given a result, cache if the user has cache enabled.
        """
        if self.settings.disable_cache == True:
            return

        # If we are using the memory cache, return from there.
        if not self.settings.disable_memory_cache:
            if name in self._cache:
                return self._cache[name]

        # Ensure cache directory exists
        utils.mkdir_p(self.settings.cache_dir)

        # prepare the path (e.g., cache_dir/package_managers.json)
        path = self.get_cache_name(name)

        # Don't write empty data
        if not result.data:
            logger.warning("No data found for result, not writing %s" % path)
            return

        # If we are using the memory cache, save to it
        if not self.settings.disable_memory_cache:
            self._cache[name] = result.data

        # We can't predict nesting, so always make directory
        utils.mkdir_p(os.path.dirname(path))
        utils.write_json(result.data, path)

    def get_cache_name(self, name):
        """
        Return a json cache entry.
        """
        return os.path.join(self.settings.cache_dir, "%s.json" % name)

    def get_cache(self, name, endpoint=None):
        """
        Given a cache name (typically matching the endpoint) retrieve if exists.
        If provided and endpoint, wrap the result with the endpoint. Otherwise,
        return the json result.
        """
        path = self.get_cache_name(name)
        if not os.path.exists(path):
            return

        # Load the cache, return as a result if it exists.
        data = utils.read_json(path)
        if data and endpoint:
            return results.Table(data, endpoint)
        elif data:
            return data

    def get_endpoint(self, name, data=None, **kwargs):
        """
        Get a named endpoint, optionally, using the cache (default)
        """
        if name not in endpoints.registry:
            names = endpoints.registry_names
            logger.exit(f"{name} is not a known endpoint. Choose from {names}")

        # Create the endpoint with any optional params
        if not data:
            endpoint = endpoints.registry[name](**kwargs, require_params=False)
            return results.Table(self.get(endpoint.url), endpoint)
        endpoint = endpoints.registry[name](**kwargs)
        return results.Table(data, endpoint)

    def check_response(self, typ, r, return_json=True, stream=False, retry=True):
        """
        Ensure the response status code is 20x
        """
        # Rate is 60/minute
        if r.status_code == 429:
            logger.info("Exceeded API limit, sleeping 1 minute.")
            time.sleep(60)
            r = self.session.send(r.request)
            return self.check_response(typ, r, return_json, stream, retry=retry)

        if r.status_code == 401:
            logger.exit("You must set CITELAG_LIBRARIES_KEY in the environment.")

        if r.status_code not in [200, 201]:
            logger.exit("Unsuccessful response: %s, %s" % (r.status_code, r.reason))

        # All data is typically json
        if return_json and not stream:
            return r.json()
        return r

    def print_response(self, r):
        """
        Print the result of a response
        """
        response = r.json()
        logger.info("%s: %s" % (r.url, json.dumps(response, indent=4)))

    def do_request(
        self,
        typ,
        url,
        data=None,
        json=None,
        headers=None,
        return_json=True,
        stream=False,
    ):
        """
        Do a request (get, post, etc)
        """
        # If we have a cached token, use it!
        headers = headers or {}
        headers.update(self.headers)

        if not self.quiet:
            logger.info("%s %s" % (typ.upper(), url))

        # The first post when you upload the model defines the flavor (regression)
        if json:
            r = requests.request(typ, url, json=json, headers=headers, stream=stream)
        else:
            r = requests.request(typ, url, data=data, headers=headers, stream=stream)
        if not self.quiet and not stream and not return_json:
            self.print_response(r)
        return self.check_response(typ, r, return_json=return_json, stream=stream)

    def post(self, url, data=None, json=None, headers=None, return_json=True):
        """
        Perform a POST request
        """
        return self.do_request(
            "post", url, data=data, json=json, headers=headers, return_json=return_json
        )

    def delete(self, url, data=None, json=None, headers=None, return_json=True):
        """
        Perform a DELETE request
        """
        return self.do_request(
            "delete",
            url,
            data=data,
            json=json,
            headers=headers,
            return_json=return_json,
        )

    def get(
        self, url, data=None, json=None, headers=None, return_json=True, stream=False
    ):
        """
        Perform a GET request
        """
        return self.do_request(
            "get",
            url,
            data=data,
            json=json,
            headers=headers,
            return_json=return_json,
            stream=stream,
        )



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from citelang.main import Client
from citelang.logger import logger
import tempfile
import os


def main(args, parser, extra, subparser):

    cli = Client(quiet=args.quiet, settings_file=args.settings_file)
    result = cli.badge(
        name=args.package[1],
        manager=args.package[0],
        use_cache=not args.no_cache,
        max_depth=args.max_depth,
        max_deps=args.max_depth,
        min_credit=args.min_credit,
        credit_split=args.credit_split,
        template=args.template,
    )

    if not args.outfile and args.template == "static":
        args.outfile = os.path.join("%s-%s.png" % (args.package[0], args.package[1]))

    elif not args.outfile:
        args.outfile = os.path.join(tempfile.mkdtemp(), "index.html")

    if os.path.exists(args.outfile) and not args.force:
        logger.exit(f"{args.outfile} exists, add --force to overwrite.")
    result.save(args.outfile)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import os
import stat
import json
import sys
import errno
import tempfile


def mkdir_p(path):
    """mkdir_p attempts to get the same functionality as mkdir -p
    :param path: the path to create.
    """
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            sys.exit("Error creating path %s, exiting." % path)


def write_file(content, filename, mode="w", exec=False):
    """
    Write content to a filename
    """
    with open(filename, mode) as filey:
        filey.writelines(content)
    if exec:
        st = os.stat(filename)
        os.chmod(filename, st.st_mode | stat.S_IEXEC)
    return filename


def write_json(json_obj, filename, mode="w"):
    """
    Write json to a filename
    """
    with open(filename, mode) as filey:
        filey.writelines(print_json(json_obj))
    return filename


def print_json(json_obj):
    """
    Print json pretty
    """
    return json.dumps(json_obj, indent=4, separators=(",", ": "))


def read_file(filename, mode="r"):
    """
    Read a file.
    """
    with open(filename, mode) as filey:
        content = filey.read()
    return content


def read_json(filename, mode="r"):
    """
    Read a json file to a dictionary.
    """
    return json.loads(read_file(filename))


def get_tmpfile(tmpdir=None, prefix=""):
    """
    Get a temporary file with an optional prefix.
    """
    # First priority for the base goes to the user requested.
    tmpdir = get_tmpdir(tmpdir)

    # If tmpdir is set, add to prefix
    if tmpdir:
        prefix = os.path.join(tmpdir, os.path.basename(prefix))

    fd, tmp_file = tempfile.mkstemp(prefix=prefix)
    os.close(fd)

    return tmp_file


def get_tmpdir(tmpdir=None, prefix="", create=True):
    """
    Get a temporary directory for an operation.
    """
    tmpdir = tmpdir or tempfile.gettempdir()
    prefix = prefix or "citelang-tmp"
    prefix = "%s.%s" % (prefix, next(tempfile._get_candidate_names()))
    tmpdir = os.path.join(tmpdir, prefix)

    if not os.path.exists(tmpdir) and create is True:
        os.mkdir(tmpdir)

    return tmpdir



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"


from citelang.logger import logger

import sys

from .base import GraphBase


class Cypher(GraphBase):
    def generate(self):
        if self.outfile == sys.stdout:
            fd = sys.stdout
        else:
            logger.info("Output will be written to %s" % self.outfile)
            fd = open(self.outfile, "w")

        fd.write("CREATE ")

        # Create graph with nodes
        next_nodes = [self.data]
        while next_nodes:
            next_node = next_nodes.pop(0)
            label = self.uids[next_node.name]
            name = "%s (%s)" % (next_node.name, round(next_node.credit, 3))
            fd.write("(%s:PACKAGE {name: '%s', label: '%s'}),\n" % (label, name, label))
            for child in next_node.children:
                next_nodes.append(child)

        # Now create links
        next_nodes = [self.data]
        text = ""
        while next_nodes:
            next_node = next_nodes.pop(0)
            label = self.uids[next_node.name]
            name = "%s (%s)" % (next_node.name, round(next_node.credit, 3))
            for i, child in enumerate(next_node.children):
                label_to = self.uids[child.name]
                text += "(%s)-[:DEPENDSON]->(%s),\n" % (label, label_to)
                next_nodes.append(child)

        text = text.strip(",\n")
        fd.write(text + ";\n")
        if self.outfile != sys.stdout:
            fd.close()



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import colour
import random

# Palettes

# https://coolors.co/a80874-ac4697-b083b9-b4c0dc-b7fdfe-8bf8c5-5ef38c-45c556-2b9720-343a1a
palettes = {
    "radish": [
        "a80874",
        "ac4697",
        "b083b9",
        "b4c0dc",
        "b7fdfe",
        "8bf8c5",
        "5ef38c",
        "45c556",
        "2b9720",
        "343a1a",
    ]
}


def get_rich_color():
    """
    Return a random color
    """
    return "color(" + str(random.choice(range(255))) + ")"


def get_rich_colors(N):
    """
    Randomly generate N colors for rich (integers)
    """
    chosen_colors = []

    for i in range(N):
        color = None
        while not color or color in chosen_colors:
            color = get_rich_color()
        chosen_colors.append(color)
    return chosen_colors


def get_random_palette():
    """
    Choose a random color palette
    """
    name = random.choice(list(palettes.keys()))
    return palettes[name]


def get_color_range(N=100):
    """
    Get a color range, size N
    """
    palette = get_random_palette()
    start_color = colour.Color("#" + palette.pop(random.choice(range(len(palette)))))
    end_color = colour.Color("#" + palette.pop(random.choice(range(len(palette)))))
    return list(start_color.range_to(end_color, N))



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from citelang.main import Client


def main(args, parser, extra, subparser):

    cli = Client(quiet=args.quiet, settings_file=args.settings_file)
    result = cli.package_managers(use_cache=not args.no_cache)

    if args.json and not args.outfile:
        result.print_json()

    elif args.outfile:
        result.save(args.outfile)

    else:
        result.table()



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"


from .base import GraphBase
from .tree import print_tree


class Console(GraphBase):
    def generate(self):
        print_tree(self.data)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

# Custom package managers not in libraries IO

import sys
from citelang.logger import logger
import requests


class PackageManager:
    """
    A package manager is a custom class (not supported by libraries io)
    """

    def __init__(self, *args, **kwargs):
        for attr in [
            "name",
            "project_count",
            "homepage",
            "color",
            "default_language",
            "default_versions",
        ]:
            if not hasattr(self, attr):
                sys.exit(
                    "Misconfigured package manager %s missing %s attribute"
                    % (self.name, attr)
                )

    # If needed, make sure endpoint data is sorted
    def info(self):
        return {
            "name": self.name,
            "project_count": self.project_count,
            "homepage": self.homepage,
            "color": self.color,
            "default_language": self.default_language,
        }

    def get_or_fail(self, url, headers=None, return_text=False):
        """
        Shared endpoint to get a url or fail.
        """
        response = requests.get(url, headers=headers)
        if response.status_code != 200:
            logger.exit("Cannot retrieve %s: %s" % (url, response.json()))
        if return_text:
            return response.text
        return response.json()

    def package(self, name, **kwargs):
        raise NotImplementedError



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from citelang.logger import logger
import citelang.utils as utils
from operator import itemgetter

import re
import os

table_template = """|Manager|Name|Credit|
|-------|----|------|
"""

template_suffix = "\n> Note that credit values are rounded and expanded (so shared dependencies are represented as one record) and may not add to 1.0. Rounded values that hit zero are removed.\n"

citation_regex = "\@([a-zA-Z0-9]+)\{(.*?)\}"
empty_content = """# Software Credit

<!--citelang start-->
<!--citelang end-->

- Generated by [CiteLang](https://github.com/vsoch/citelang)
"""


class Parser:
    """
    A parser reads in a markdown file, finds software references, and generates
    a software graph per the user preferences.
    """

    def __init__(self, filename: str = None):
        self.data = {}
        self.libs = []
        self.round_by = 3
        self.filename = None
        if filename:
            self.filename = os.path.abspath(filename)
            if not os.path.exists(self.filename):
                logger.exit(f"{filename} does not exist")

            self.content = utils.read_file(self.filename)
        else:
            self.content = empty_content

    @property
    def start_block(self):
        return "<!--citelang start-->"

    @property
    def end_block(self):
        return "<!--citelang end-->"

    def check(self):
        """
        Checks for citelang.
        """
        if self.start_block not in self.content:
            logger.exit(
                "Cannot find %s, ensure it is present where you want the citelang table."
                % self.start_block
            )
        if self.end_block not in self.content:
            logger.exit(
                "Cannot find %s, ensure it is present where you want the citelang table."
                % self.end_block
            )

    def parse(self):
        """
        Given a markdown file, return a list of parse packages and versions.
        """
        self.check()
        for match in re.findall(citation_regex, self.content):
            if len(match) != 2:
                logger.warning(
                    "found malformed citation reference %s, skipping." % match
                )
                continue
            args = {
                y[0]: y[1]
                for y in [x.strip().split("=") for x in match[-1].split(",")]
                if len(y) == 2
            }
            self.libs.append({"manager": match[0], **args})

    def prepare_table(self, roots):
        """
        Given a set of roots, prepare table data.
        """
        # Generate the table with multiple roots - flatten out credit
        table = {}

        # Multiplier for credit depending on total packages
        splitby = 1 / len(roots)
        for lib, root in roots.items():
            manager = lib.split(":")[0]
            for node in root.iternodes():
                if manager not in table:
                    table[manager] = {}
                if node.name not in table[manager]:
                    table[manager][node.name] = {
                        "credit": 0,
                        "url": node.obj.homepage,
                    }
                table[manager][node.name]["credit"] += node.credit * splitby

        # Add listing of packages and dependencies to parser
        self.data = table
        self.round_by = root.round_by
        return table

    def add_lib(self, manager, name, **args):
        """
        Manually add a library (e.g., not reading from a pre-existing file)
        """
        self.libs.append({"manager": manager, "name": name, **args})

    def render(self):
        """
        Render final file!
        """
        markdown = table_template

        # Sort from least to greatest
        listing = []
        for manager, packages in self.data.items():
            for package, meta in packages.items():
                listing.append((manager, package, meta["credit"], meta["url"]))

        listing = sorted(listing, key=itemgetter(2), reverse=True)
        for (manager, package, credit, url) in listing:
            credit = round(credit, self.round_by)
            if credit == 0:
                logger.warning("Rounded credit for %s is 0, skipping." % package)
                continue
            if url:
                package = "[%s](%s)" % (package, url)
            markdown += "|%s|%s|%s|\n" % (
                manager,
                package,
                credit,
            )

        render = []
        lines = self.content.split("\n")
        while lines:
            line = lines.pop(0)
            if self.start_block in line:
                while self.end_block not in line:
                    line = lines.pop(0)
                # When we get here we have the end block
                render += (
                    [self.start_block]
                    + markdown.split("\n")
                    + [template_suffix, self.end_block]
                )
            else:
                render.append(line)
        return "\n".join(render)



#!/usr/bin/env python

__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import citelang
from citelang.logger import setup_logger
import argparse
import sys
import os


def get_parser():
    parser = argparse.ArgumentParser(
        description="CiteLang",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # Global Variables
    parser.add_argument(
        "--debug",
        dest="debug",
        help="use verbose logging to debug.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--quiet",
        dest="quiet",
        help="suppress additional output.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--no-cache",
        dest="no_cache",
        help="disable using the cache for any particular command.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--settings-file",
        dest="settings_file",
        help="custom path to settings file.",
    )

    parser.add_argument(
        "--version",
        dest="version",
        help="show software version.",
        default=False,
        action="store_true",
    )

    description = "actions"
    subparsers = parser.add_subparsers(
        help="actions",
        title="actions",
        description=description,
        dest="command",
    )

    # print version and exit
    subparsers.add_parser("version", description="show software version")

    # List package managers available
    ls = subparsers.add_parser(
        "list", description="list package managers available to derive citations from."
    )
    ls.add_argument("--json", help="output json", default=False, action="store_true")
    ls.add_argument("--all", help="output json", default=False, action="store_true")
    ls.add_argument("--outfile", "-o", help="write content to output file")

    # Cache control
    cache = subparsers.add_parser("cache", description="cache control")
    cache.add_argument(
        "--clear", help="clear the cache", default=False, action="store_true"
    )

    # Get a package or dependencies
    pkg = subparsers.add_parser(
        "package",
        description="list package managers available to derive citations from.",
    )
    deps = subparsers.add_parser("deps", description="list dependencies for a package.")

    badge = subparsers.add_parser(
        "badge", description="Generate an html (svg-based) badge."
    )
    badge.add_argument(
        "--template",
        "-t",
        help="Template to use for badge",
        choices=["treemap", "sunburst", "static"],
        default="static",
    )

    for command in [cache, badge]:
        command.add_argument(
            "--force",
            help="do not ask for confirmation",
            default=False,
            action="store_true",
        )

    graph = subparsers.add_parser(
        "graph", description="generate a graph for some package dependencies."
    )
    graph.add_argument(
        "--fmt", "-f", help="format of the graph (unset defaults to console print)"
    )
    credit = subparsers.add_parser(
        "credit", description="calculate dependency credit for a package."
    )
    gen = subparsers.add_parser(
        "gen", description="Generate a credit markdown file for a package of choice."
    )
    gen.add_argument("--outfile", "-o", help="Save to an output markdown file.")

    render = subparsers.add_parser(
        "render", description="render a credit tree into your markdown."
    )
    render.add_argument("filename", help="Markdown file to render software table into.")
    render.add_argument("--outfile", "-o", help="Save to an output json file.")

    for command in [pkg, deps, graph, credit, badge, gen]:
        command.add_argument(
            "package", help="package manager and name to parse", nargs=2
        )

    for command in [pkg, deps, graph, credit, badge]:
        command.add_argument(
            "--json",
            dest="json",
            help="print output to json.",
            default=False,
            action="store_true",
        )
        command.add_argument("--outfile", "-o", help="Save to an output json file.")

    for command in [graph, credit, badge, render, gen]:
        command.add_argument(
            "--max-depth", help="maximum depth to parse tree (default is unset)"
        )
        command.add_argument(
            "--max-deps",
            help="maximum number of dependencies to include (default is unset)",
        )
        command.add_argument(
            "--min-credit",
            help="minimum credit to allocate (and stop when lower)",
            type=float,
            default=0.01,
        )
        command.add_argument(
            "--credit-split",
            help="credit to allocate to main package and 1-split to dependents",
            type=float,
            default=0.5,
        )

    # Local shell with client loaded
    shell = subparsers.add_parser(
        "shell",
        description="shell into a Python session with a client.",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    shell.add_argument(
        "--interpreter",
        "-i",
        dest="interpreter",
        help="python interpreter",
        choices=["ipython", "python", "bpython"],
        default="ipython",
    )

    # Custom config
    config = subparsers.add_parser(
        "config",
        description="update configuration settings. Use set or get to see or set information.",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    config.add_argument(
        "--central",
        dest="central",
        help="make edits to the central config file.",
        default=False,
        action="store_true",
    )

    config.add_argument(
        "params",
        nargs="*",
        help="""Set or get a config value, edit the config, add or remove a list variable, or create a user-specific config.
citelang config set key:value
citelang config get key
citelang edit
citelang config inituser""",
        type=str,
    )
    return parser


def run():

    parser = get_parser()

    def help(return_code=0):
        version = citelang.__version__

        print("\nCiteLang Client v%s" % version)
        parser.print_help()
        sys.exit(return_code)

    # If the user didn't provide any arguments, show the full help
    if len(sys.argv) == 1:
        help()

    # If an error occurs while parsing the arguments, the interpreter will exit with value 2
    args, extra = parser.parse_known_args()

    if args.debug is True:
        os.environ["MESSAGELEVEL"] = "DEBUG"

    # Show the version and exit
    if args.command == "version" or args.version:
        print(citelang.__version__)
        sys.exit(0)

    setup_logger(
        quiet=args.quiet,
        debug=args.debug,
    )

    # retrieve subparser (with help) from parser
    helper = None
    subparsers_actions = [
        action
        for action in parser._actions
        if isinstance(action, argparse._SubParsersAction)
    ]
    for subparsers_action in subparsers_actions:
        for choice, subparser in subparsers_action.choices.items():
            if choice == args.command:
                helper = subparser
                break

    if args.command == "badge":
        from .badge import main
    elif args.command == "cache":
        from .cache import main
    elif args.command == "config":
        from .config import main
    elif args.command == "credit":
        from .credit import main
    elif args.command == "deps":
        from .deps import main
    elif args.command == "gen":
        from .gen import main
    elif args.command == "graph":
        from .graph import main
    elif args.command == "render":
        from .render import main
    elif args.command == "shell":
        from .shell import main
    elif args.command == "list":
        from .listing import main
    elif args.command == "package":
        from .package import main

    # Pass on to the correct parser
    return_code = 0
    try:
        main(args=args, parser=parser, extra=extra, subparser=helper)
        sys.exit(return_code)
    except UnboundLocalError:
        return_code = 1

    help(return_code)


if __name__ == "__main__":
    run()
